module solver
  implicit none

  integer, parameter :: idens=1, iener=4, imomx=2, imomy=3, nvars=4
  real, parameter :: gamma=5./3.

  public :: initialconditions, timestep
  private:: cfl, xytranspose, xsweep, tvd1d, hydroflux

contains

subroutine timestep(u,ut,dt)
    real, dimension(:,:,:), intent(INOUT) :: u[*]
    real, dimension(:,:,:), intent(INOUT) :: ut[*]
    real, intent(OUT) :: dt

    real, save :: localdt[*]
    integer :: i, nx, ny
    integer :: me, num, leftneigh, rightneigh

    me = this_image()
    num = num_images()

    leftneigh = me-1
    if (leftneigh < 1) leftneigh = num
    rightneigh = me + 1
    if (rightneigh > num) rightneigh = 1
    
    nx = size(u,2)
    ny = size(u,3)

! each find our local dt
    localdt=0.5*cfl(u)
    sync all()
! find min of these dts
    dt = 1.e9
    do i=1,num_images()
      dt = min(dt,localdt[i])
    enddo

! the x sweep
    call xsweep(u,dt,.false.)
!! exchange boundary conditions in x-dirn
    call xytranspose(ut,u)

    call xsweep(ut,dt,.true.)
    sync images( (/ leftneigh, me, rightneigh /) )
!! exchange boundary conditions in x-dirn
    ut(:,:,1)  = ut(:,:,nx-1)[leftneigh]
    ut(:,:,nx) = ut(:,:,2)  [rightneigh]
    sync images( (/ leftneigh, me, rightneigh /) )

    call xsweep(ut,dt,.true.)
!! exchange boundary conditions in x-dirn

! 2nd x sweep
    call xytranspose(u,ut)
    call xsweep(u,dt,.false.)

!! exchange boundary conditions in x-dirn
    sync images( (/ leftneigh, me, rightneigh /) )
    u(:,1,:)  = u(:,nx-1,:)[leftneigh]
    u(:,nx,:) = u(:,2,:)  [rightneigh]
    sync images( (/ leftneigh, me, rightneigh /) )
end subroutine timestep


real function cfl(u)
   implicit none
   real, intent(IN), target :: u(:,:,:)
   real, dimension(size(u,2),size(u,3)) :: v, p
   real, dimension(:,:), pointer :: dens, momx, momy, ener
   real :: c

   dens => u(idens,:,:)
   momx => u(imomx,:,:)
   momy => u(imomy,:,:)
   ener => u(iener,:,:)

   v = abs(momx/dens)
   where (v < abs(momy/dens))
      v = abs(momy/dens)
   endwhere

   p = (ener - 0.5*(momx**2+momy**2)/dens)*(gamma-1.)
   c = maxval(v+sqrt(abs((gamma*p))/dens))
   
   cfl=1./c
end function cfl

pure subroutine initialconditions(u,start)
  implicit none
  real, intent(OUT)   :: u(:,:,:)
  integer, intent(IN), optional :: start

  integer :: nx,ny
  integer :: i,j
  real, parameter :: backgrounddens=1., projdens=50., projvel=4.
  real :: r(size(u,2), size(u,3))

  integer :: shift
  if (present(start)) then
    shift = start
  else
    shift = 0
  endif
 
  nx = size(u,2)
  ny = size(u,3)
 
  forall (i=1:nx,j=1:ny)
      r(i,j) =sqrt((j*1-ny/2.)**2. + (i*1.-nx/2.+shift)**2)
  endforall

  where (r < 0.1*sqrt(nx*nx*1.+ny*ny))
    u(idens,:,:) =projdens
    u(imomx,:,:) =projdens*projvel
    u(imomy,:,:) =0
    u(iener,:,:) =0.5*(projdens*projvel*projvel)+1./((gamma-1.)*projdens)
  elsewhere
    u(idens,:,:) =backgrounddens
    u(imomx,:,:) =0.
    u(imomy,:,:) =0.
    u(iener,:,:) =1./((gamma-1.)*backgrounddens)
  endwhere

end subroutine initialconditions

subroutine xsweep(u,dt,periodic)
  implicit none
  real, intent(INOUT), dimension(:,:,:) :: u
  real, intent(IN) :: dt
  logical, intent(in) :: periodic
  integer :: j

  do j=1,size(u,3)
     call tvd1d(u(:,:,j),dt,periodic)
  enddo
end subroutine xsweep

function shift(arr, sh, dimen, periodic)
  real, dimension(:,:), intent(in) :: arr
  real, dimension(size(arr,1),size(arr,2)) :: shift
  integer, intent(in) :: sh, dimen
  logical, intent(in) :: periodic

  if (periodic) then
    shift = cshift(arr, sh, dimen)
  else
    if (sh > 0) then
        shift = eoshift(arr, sh, boundary=arr(:,size(arr,1)), dim=dimen)
    else
        shift = eoshift(arr, sh, boundary=arr(:,1), dim=dimen)
    endif
  endif
end function shift

subroutine tvd1d(u,dt,periodic)
  real, intent(IN) :: dt
  real, dimension(:,:), intent(INOUT) :: u
  logical, intent(in), optional :: periodic

  real, dimension(size(u,1),size(u,2)) :: v, u1, wr, wl
  real, dimension(size(u,1),size(u,2)) :: dfrp, dfrm, dflm, dflp
  real, dimension(size(u,1),size(u,2)) :: dfl, dfr
  real, dimension(size(u,1),size(u,2)) :: flux
  logical :: period
  real :: c

  period = .true.
  if (periodic) period=periodic

  call hydroflux(v,c,u)
  wr = u+v
  wl = u-v
  flux=c*(wr - shift(wl,1,2,period))/2.
  u1 = u - (flux - shift(flux,-1,2,period))*dt/2.

  call hydroflux(v,c,u1)
  wr = u1+v
  wl = u1-v

  dfrp = c*(shift(wr,1,2,period)-wr)/2.
  dfrm = c*(wr-shift(wr,-1,2,period))/2.
  where (dfrp*dfrm > 0) 
     dfr = 2*dfrp*dfrm/(dfrp+dfrm)
  elsewhere
     dfr = 0
  endwhere
  
  dflp = c*(shift(wl,1,2,period)-shift(wl,2,2,period))/2.
  dflm = c*(wl-shift(wl,1,2,period))/2.
  where (dflp*dflm > 0) 
     dfl = 2*dflp*dflm/(dflp+dflm)
  elsewhere
     dfl = 0
  endwhere
  
  flux = (c*(wr-shift(wl,1,2,period)) + (dfr-dfl))/2.
  u = u-(flux-shift(flux,-1,2,period))*dt
  return
end subroutine tvd1d

subroutine hydroflux(v,c,u)
  implicit none
  real, intent(OUT)    :: c
  real, dimension(:,:), intent(IN), target :: u
  real, dimension(:,:), intent(OUT) :: v
  real, dimension(size(u,2)) :: vx, p
  real, dimension(:), pointer :: momx, momy, ener, dens

  momx => u(imomx,:)
  momy => u(imomy,:)
  ener => u(iener,:)
  dens => u(idens,:)

  c  = 0.
  vx = momx/dens
  p  = (ener- 0.5*(momx**2+momy**2)/(dens))*(gamma-1)

  v(idens,:)=momx
  v(imomx,:)=momx*vx + p
  v(imomy,:)=momy*vx
  v(iener,:)=(ener+p)*vx

  c=maxval(abs(vx)+sqrt(abs(gamma*p/dens)))
  if (c > 0) v = v/c
end subroutine hydroflux

pure subroutine xytranspose(ut,u)
  implicit none
  real, dimension(:,:,:), intent(IN) :: u
  real, dimension(:,:,:), intent(OUT) :: ut

  ut(idens,:,:)=transpose(u(idens,:,:))
  ut(imomx,:,:)=transpose(u(imomy,:,:))
  ut(imomy,:,:)=transpose(u(imomx,:,:))
  ut(iener,:,:)=transpose(u(iener,:,:))
end subroutine xytranspose

end module solver
