module solver

  implicit none

  integer, parameter :: idens=1, iener=4, imomx=2, imomy=3
  real, parameter :: gamma=5./3.

  public :: initialconditions, timestep
  private:: cfl, xytranspose, xsweep, tvd1d, hydroflux

contains

pure subroutine timestep(u,nx,ny,dt)
    integer, intent(IN) :: nx, ny
    real, dimension(4,nx,ny), intent(INOUT) :: u
    real, intent(OUT) :: dt

    real, dimension(4,nx,ny) :: ut

    dt=0.5*cfl(u,nx,ny)
! the x sweep
    call xsweep(u,nx,ny,dt)
! the y sweeps
    call xytranspose(ut,u)
    call xsweep(ut,ny,nx,dt)
    call xsweep(ut,ny,nx,dt)
! 2nd x sweep
    call xytranspose(u,ut)
    call xsweep(u,nx,ny,dt)
end subroutine timestep

pure real function cfl(u,nx,ny)
   implicit none
   integer, intent(IN) :: nx, ny
   real, intent(IN) :: u(4,nx,ny)

   real :: v,p,c
   integer i,j
   
   c = 0.
   do j=1,ny
      do i=1,nx
         v=max(abs(u(imomx,i,j)),abs(u(imomy,i,j)))/u(idens,i,j)
         p=(u(iener,i,j)-(u(imomx,i,j)**2+u(imomy,i,j)**2)/ &
                         u(idens,i,j)/2)*(gamma-1)
         c=max(c,v+sqrt(abs((gamma*p)/u(idens,i,j))))
      enddo
   enddo
   cfl=1./c
end function cfl

pure subroutine initialconditions(u,nx,ny)
  implicit none
  integer, intent(IN) :: nx,ny
  real, intent(OUT)   :: u(4,nx,ny)

  integer :: i,j
  real, parameter :: backgrounddens=1., projdens=50., projvel=4.
  real :: r

  xloop: do i=1,nx
    yloop: do j=1,ny
      u(imomy,i,j)=0
      r = sqrt((j*1.0-ny/2.)**2 + (i*1.0-nx/2.)**2)
      if (r < 0.1*sqrt(nx*nx*1.+ny*ny*1.)) then
          u(idens,i,j) = projdens
          u(imomx,i,j) = projdens*projvel
          u(iener,i,j) =  0.5*(projdens*projvel*projvel) + (1./((gamma-1.)*projdens))
      else
          u(idens,i,j) = backgrounddens
          u(imomx,i,j) = backgrounddens*0.
          u(iener,i,j) =  0.5*(backgrounddens*0.*0.) + &
                          (1./((gamma-1.)*backgrounddens))
      endif
    enddo yloop
  enddo xloop
  return
end subroutine initialconditions

pure subroutine xsweep(u,nx,ny,dt)
  implicit none
  integer, intent(IN) :: nx,ny
  real, intent(INOUT), dimension(4,nx,ny) :: u
  real, intent(IN) :: dt
  integer :: j

  do j=1,ny
     call tvd1d(u(1,1,j),nx,dt)
  enddo
end subroutine xsweep

pure subroutine tvd1d(u,n,dt)
  integer, intent(IN) :: n
  real, intent(IN) :: dt
  real, dimension(4,n), intent(INOUT) :: u

  real, dimension(4,n) :: v, u1, wr, wl
  real :: dfrp, dfrm, dflm, dflp
  real :: dfl(4,n), dfr(4,n)
  real :: flux(4,n)
  real :: c
  integer :: i, var
  integer :: ilshift, irshift, irrshift

  call hydroflux(v,c,u)
  wr = u+v
  wl = u-v
  flux=c*(wr - cshift(wl,1,dim=2))/2.
  u1 = u - (flux - cshift(flux,-1,dim=2))*dt/2.

  call hydroflux(v,c,u1)
  wr = u1+v
  wl = u1-v
  
  do i=1,n
      irshift=mod(n+(i-1)+1,n)+1
      ilshift=mod(n+(i-1)-1,n)+1
      do var=1,4
          dfrp = c*(wr(var,irshift)-wr(var,i))/2.
          dfrm = c*(wr(var,i)-wr(var,ilshift))/2.
          dfr(var,i) =0.
          if (dfrp*dfrm > 0) then
             dfr(var,i)=2*dfrp*dfrm/(dfrp+dfrm)
          endif
      enddo
  enddo
  do i=1,n
      irshift=mod(n+(i-1)+1,n)+1
      irrshift=mod(n+(i-1)+2,n)+1
      do var=1,4
          dflp = c*(wl(var,irshift)-wl(var,irrshift))/2.
          dflm = c*(wl(var,i)-wl(var,irshift))/2.
          dfl(var,i) =0.
          if (dflp*dflm > 0) then
             dfl(var,i)=2*dflp*dflm/(dflp+dflm)
          endif
      enddo
  enddo
  do i=1,n
     irshift=mod(n+(i-1)+1,n)+1
     do var=1,4
         flux(var,i)=(c*wr(var,i)-c*wl(var,irshift)+ &
                    (dfr(var,i)-dfl(var,i)))/2.
     enddo
  enddo
  do i=1,n
     ilshift=mod(n+(i-1)-1,n)+1
     do var=1,4
        u(var,i)=u(var,i)-(flux(var,i)-flux(var,ilshift))*dt
     enddo
  enddo
  return
end subroutine tvd1d

pure subroutine hydroflux(v,c,u)
  implicit none
  real, intent(OUT)    :: c
  real, dimension(:,:), intent(IN) :: u
  real, dimension(:,:), intent(OUT) :: v
  real, dimension(size(u,2)) :: vx, p

  c  = 0.
  vx = u(imomx,:)/u(idens,:)
  p  = u(iener,:)-(u(imomx,:)**2+u(imomy,:)**2)/u(idens,:)/2
  p  = p*(gamma-1.)

  v(idens,:)=u(imomx,:)
  v(imomx,:)=u(imomx,:)*vx + p
  v(imomy,:)=u(imomy,:)*vx
  v(iener,:)=(u(iener,:)+p)*vx

  c=maxval(abs(vx)+sqrt(abs(gamma*p/u(idens,:))))
  if (c > 0) v = v/c
end subroutine hydroflux

pure subroutine xytranspose(ut,u)
  implicit none
  real, dimension(:,:,:), intent(IN) :: u
  real, dimension(:,:,:), intent(OUT) :: ut

  ut(idens,:,:)=transpose(u(idens,:,:))
  ut(imomx,:,:)=transpose(u(imomy,:,:))
  ut(imomy,:,:)=transpose(u(imomx,:,:))
  ut(iener,:,:)=transpose(u(iener,:,:))
end subroutine xytranspose

end module solver
