module solver

  implicit none

  integer, parameter :: idens=1, iener=4, imomx=2, imomy=3, nvars=4
  real, parameter :: gamma=5./3.

  public :: initialconditions, timestep
  private:: cfl, xytranspose, xsweep, tvd1d, hydroflux

contains

pure subroutine timestep(u,nx,ny,dt)
    integer, intent(IN) :: nx, ny
    real, dimension(nvars,nx,ny), intent(INOUT) :: u
    real, intent(OUT) :: dt

    real, dimension(nvars,nx,ny) :: ut

    dt=0.5*cfl(u,nx,ny)
! the x sweep
    call xsweep(u,nx,ny,dt)
! the y sweeps
    call xytranspose(ut,u,nx,ny)
    call xsweep(ut,ny,nx,dt)
    call xsweep(ut,ny,nx,dt)
! 2nd x sweep
    call xytranspose(u,ut,ny,nx)
    call xsweep(u,nx,ny,dt)
end subroutine timestep

pure real function cfl(u,nx,ny)
   implicit none
   integer, intent(IN) :: nx, ny
   real, intent(IN) :: u(nvars,nx,ny)

   real :: v,p,c
   integer i,j
   
   c = 0.
   do j=1,ny
      do i=1,nx
         v=max(abs(u(imomx,i,j)),abs(u(imomy,i,j)))/u(idens,i,j)
         p=(u(iener,i,j)-(u(imomx,i,j)**2+u(imomy,i,j)**2)/ &
                         u(idens,i,j)/2)*(gamma-1)
         c=max(c,v+sqrt(abs((gamma*p)/u(idens,i,j))))
      enddo
   enddo
   cfl=1./c
end function cfl

pure subroutine initialconditions(u,nx,ny)
  implicit none
  integer, intent(IN) :: nx,ny
  real, intent(OUT)   :: u(nvars,nx,ny)

  integer :: i,j
  real, parameter :: backgrounddens=1., projdens=50., projvel=4.
  real :: r

  xloop: do i=1,nx
    yloop: do j=1,ny
      u(imomy,i,j)=0
      r = sqrt((j*1.0-ny/2.)**2 + (i*1.0-nx/2.)**2)
      if (r < 0.1*sqrt(nx*nx*1.+ny*ny*1.)) then
          u(idens,i,j) = projdens
          u(imomx,i,j) = projdens*projvel
          u(iener,i,j) =  0.5*(projdens*projvel*projvel) + (1./((gamma-1.)*projdens))
      else
          u(idens,i,j) = backgrounddens
          u(imomx,i,j) = backgrounddens*0.
          u(iener,i,j) =  0.5*(backgrounddens*0.*0.) + &
                          (1./((gamma-1.)*backgrounddens))
      endif
    enddo yloop
  enddo xloop
  return
end subroutine initialconditions

pure subroutine xsweep(u,nx,ny,dt)
  implicit none
  integer, intent(IN) :: nx,ny
  real, intent(INOUT), dimension(nvars,nx,ny) :: u
  real, intent(IN) :: dt
  integer :: j

  do j=1,ny
     call tvd1d(u(1,1,j),nx,dt)
  enddo
end subroutine xsweep

pure subroutine tvd1d(u,n,dt)
  integer, intent(IN) :: n
  real, intent(IN) :: dt
  real, dimension(nvars,n), intent(INOUT) :: u

  real, dimension(nvars,n) :: v, u1, wr, wl
  real :: dfrp, dfrm, dflm, dflp
  real :: dfl(nvars,n), dfr(nvars,n)
  real :: flux(nvars,n)
  real :: c
  integer :: i, var
  integer :: ilshift, irshift, irrshift

  call hydroflux(v,c,u,n)
  do i=1,n
      do var=1,nvars
          wr(var,i)=u(var,i)+v(var,i)
          wl(var,i)=u(var,i)-v(var,i)
      enddo
  enddo
  do i=1,n
      irshift=mod(n+(i-1)+1,n)+1
      do var=1,nvars
          flux(var,i)=c*(wr(var,i)-wl(var,irshift))/2.
      enddo
  enddo
  do i=1,n
      ilshift=mod(n+(i-1)-1,n)+1
      do var=1,nvars
          u1(var,i)=u(var,i)-(flux(var,i)-flux(var,ilshift))*dt/2.
      enddo
  enddo

  call hydroflux(v,c,u1,n)
  do i=1,n
      do var=1,nvars
          wr(var,i)=u1(var,i)+v(var,i)
          wl(var,i)=u1(var,i)-v(var,i)
      enddo
  enddo
  do i=1,n
      irshift=mod(n+(i-1)+1,n)+1
      ilshift=mod(n+(i-1)-1,n)+1
      do var=1,nvars
          dfrp = c*(wr(var,irshift)-wr(var,i))/2.
          dfrm = c*(wr(var,i)-wr(var,ilshift))/2.
          dfr(var,i) =0.
          if (dfrp*dfrm > 0) then
             dfr(var,i)=2*dfrp*dfrm/(dfrp+dfrm)
          endif
      enddo
  enddo
  do i=1,n
      irshift=mod(n+(i-1)+1,n)+1
      irrshift=mod(n+(i-1)+2,n)+1
      do var=1,nvars
          dflp = c*(wl(var,irshift)-wl(var,irrshift))/2.
          dflm = c*(wl(var,i)-wl(var,irshift))/2.
          dfl(var,i) =0.
          if (dflp*dflm > 0) then
             dfl(var,i)=2*dflp*dflm/(dflp+dflm)
          endif
      enddo
  enddo
  do i=1,n
     irshift=mod(n+(i-1)+1,n)+1
     do var=1,nvars
         flux(var,i)=(c*wr(var,i)-c*wl(var,irshift)+ &
                    (dfr(var,i)-dfl(var,i)))/2.
     enddo
  enddo
  do i=1,n
     ilshift=mod(n+(i-1)-1,n)+1
     do var=1,nvars
        u(var,i)=u(var,i)-(flux(var,i)-flux(var,ilshift))*dt
     enddo
  enddo
  return
end subroutine tvd1d

pure subroutine hydroflux(v,c,u,n)
  implicit none
  integer, intent(IN) :: n
  real, intent(OUT)    :: c
  real, dimension(n) :: vx, p
  real, dimension(nvars,n), intent(OUT) :: v
  real, dimension(nvars,n), intent(IN) :: u
  integer :: i

  c = 0.
  do i=1,n
    vx(i)=u(imomx,i)/u(idens,i)
    p(i)=u(iener,i)-(u(imomx,i)**2+u(imomy,i)**2)/u(idens,i)/2
    p(i)=p(i)*(gamma-1)
    v(idens,i)=u(imomx,i)
    v(imomx,i)=u(imomx,i)*vx(i)+p(i)
    v(imomy,i)=u(imomy,i)*vx(i)
    v(iener,i)=(u(iener,i)+p(i))*vx(i)
    c=max(c, abs(vx(i))+sqrt(abs(gamma*p(i)/u(idens,i))))
  enddo
  if (c > 0) then
     do i=1,n
        v(idens,i) = v(idens,i)/c
        v(imomx,i) = v(imomx,i)/c
        v(imomy,i) = v(imomy,i)/c
        v(iener,i) = v(iener,i)/c
     enddo
  endif
end subroutine hydroflux

pure subroutine xytranspose(ut,u,nx,ny)
  implicit none
  integer, intent(IN) :: nx,ny
  real, dimension(nvars,nx,ny), intent(IN) :: u
  real, dimension(nvars,nx,ny), intent(OUT) :: ut

  integer :: i,j

  do j=1,ny
     do i=1,nx
        ut(idens,j,i)=u(idens,i,j)
        ut(imomx,j,i)=u(imomy,i,j)
        ut(imomy,j,i)=u(imomx,i,j)
        ut(iener,j,i)=u(iener,i,j)
     enddo
  enddo
end subroutine xytranspose

end module solver
